/*
 * @OSF_COPYRIGHT@
 * COPYRIGHT NOTICE
 * Copyright (c) 1990, 1991, 1992, 1993, 1996 Open Software Foundation, Inc.
 * ALL RIGHTS RESERVED (DCE).  See the file named COPYRIGHT.DCE in the
 * src directory for the full copyright text.
 */
/*
 * HISTORY
 * $Log: cnid.c,v $
 * Revision 1.1.33.2  1996/02/18  00:01:49  marty
 * 	Update OSF copyright years
 * 	[1996/02/17  22:54:33  marty]
 *
 * Revision 1.1.33.1  1995/12/08  00:17:16  root
 * 	Submit OSF/DCE 1.2.1
 * 	[1995/12/07  23:57:34  root]
 * 
 * Revision 1.1.31.1  1994/01/21  22:34:17  cbrooks
 * 	RPC Code Cleanyp - Initial Submission
 * 	[1994/01/21  21:54:07  cbrooks]
 * 
 * Revision 1.1.2.3  1993/01/03  22:58:41  bbelch
 * 	Embedding copyright notice
 * 	[1993/01/03  19:59:28  bbelch]
 * 
 * Revision 1.1.2.2  1992/12/23  20:15:20  zeliff
 * 	Embedding copyright notice
 * 	[1992/12/23  15:31:13  zeliff]
 * 
 * Revision 1.1  1992/01/19  03:05:53  devrcs
 * 	Initial revision
 * 
 * $EndLog$
 */
/*
**  Copyright (c) 1989 by
**      Hewlett-Packard Company, Palo Alto, Ca. &
**      Digital Equipment Corporation, Maynard, Mass.
**
**
**  NAME
**
**      cnid.c
**
**  FACILITY:
**
**      Remote Procedure Call (RPC)
**
**  ABSTRACT:
**
**  The Local Identifier Service.
**
**
*/

#include <commonp.h>    /* Common declarations for all RPC runtime */
#include <com.h>        /* Common communications services */
#include <comprot.h>    /* Common protocol services */
#include <cnp.h>        /* NCA Connection private declarations */
#include <cnid.h>

/*
 * INTERNAL variables.
 */

INTERNAL unsigned16     seqnum;

/***********************************************************************/


/*
**++
**
**  ROUTINE NAME:       rpc__cn_init_seqnum
**
**  SCOPE:              PRIVATE - declared in cnid.h
**
**  DESCRIPTION:
**
**
**  This routine initializes the global sequence number cell and
**  corresponding mutex.
**
**  A sequence number generated by extracting the lower sixteen bits
**  of system time in seconds.
**
**  INPUTS:             none
**
**  INPUTS/OUTPUTS:     none
**
**  OUTPUTS:            none
**
**  IMPLICIT INPUTS:    none
**
**  The system time.
**
**  IMPLICIT OUTPUTS:   none
**
**  The static variable seqnum will contain the lower 16
**  bits of the system time in seconds.
**
**  FUNCTION VALUE:     none
**
**  SIDE EFFECTS:       none
**
**
**
**
**--
**/

PRIVATE void rpc__cn_init_seqnum (void)

{

    /*
     * Determine the system time in seconds via the time() C runtime library
     * call. Use the lower 16 bits of this for the sequence number.
     */
    seqnum = ((unsigned32) time (NULL)) & 0xffff;
}


/*
**++
**
**  ROUTINE NAME:       rpc__cn_gen_local_id
**
**  SCOPE:              PRIVATE - declared in cnid.h
**
**  DESCRIPTION:
**
**  This routine creates a new local identifier. The index portion
**  of the identifier is taken from the index argument passed by the
**  caller.
**
**  INPUTS:
**
**      index           Index value to be stored in the local
**                      identifier being created.
**
**  INPUTS/OUTPUTS:     none
**
**  OUTPUTS:
**
**      lcl_id          The created local identifier.
**
**  IMPLICIT INPUTS:    none
**
**  IMPLICIT OUTPUTS:   none
**
**  FUNCTION VALUE:     none
**
**  SIDE EFFECTS:
**
**  The sequence number is incremented.
**
**--
**/

PRIVATE void rpc__cn_gen_local_id 
#ifdef _DCE_PROTO_
(
    unsigned32              index,
    rpc_cn_local_id_t       *lcl_id
)
#else
(index, lcl_id)
unsigned32              index;
rpc_cn_local_id_t       *lcl_id;
#endif
{

    /*
     * Increment it so that the local id being generated will be different that
     * the last.
     */
    seqnum++;

    /*
     * Since zero is defined as being an invalid local id we never want to
     * generate a local id with a sequence number part of zero (since the index
     * given may be zero.
     */
    if (seqnum == 0)
    {
        seqnum = 1;
    }

    /*
     * Copy the sequence number into the returned local id.
     */
    lcl_id->parts.id_seqnum = seqnum;

    /*
     * Put the caller provided index into the returned local id.
     */
    lcl_id->parts.id_index = (unsigned16) index;
}
